class InvalidCredentials(Exception):
    pass


class InvalidPeriodicity(Exception):
    pass


class InsufficientFunds(Exception):
    pass


class InvalidSort(Exception):
    pass


class InvalidType(Exception):
    pass


class PeerberryException(Exception):
    pass
